@@ -1,7 +1,7 @@
 # -*-Perl-*-
-# $Id: Sybase.pm,v 1.41 2002/10/23 21:58:07 mpeppler Exp $
+# $Id: Sybase.pm,v 1.72 2004/09/22 06:36:25 mpeppler Exp $
 
-# Copyright (c) 1996-2002   Michael Peppler
+# Copyright (c) 1996-2004   Michael Peppler
 #
 #   You may distribute under the terms of either the GNU General Public
 #   License or the Artistic License, as specified in the Perl README file.
@@ -14,16 +14,21 @@
     use DBI ();
     use DynaLoader ();
     use Exporter ();
+
+    use Sys::Hostname ();
+
     @ISA = qw(DynaLoader Exporter);
 
     @EXPORT = qw(CS_ROW_RESULT CS_CURSOR_RESULT CS_PARAM_RESULT
 		 CS_STATUS_RESULT CS_MSG_RESULT CS_COMPUTE_RESULT);
 
 
-    $VERSION = '0.95';
-    my $Revision = substr(q$Revision: 1.41 $, 10);
+    $hostname = Sys::Hostname::hostname();
+    $init_done = 0;
+    $VERSION = '1.04.8';
+    my $Revision = substr(q$Revision: 1.72 $, 10);
 
-    require_version DBI 1.02;
+    require_version DBI 1.30;
 
     bootstrap DBD::Sybase $VERSION;
 
@@ -33,17 +38,6 @@
     $errstr = '';
     $sqlstate = "00000";
 
- if(0) {
-    my %syb_api = (
-	'nsql' => ['db', { U =>[2,0,'$statement [, $type [, $callback ] ]'] } ],
-			);
-    foreach my $method (keys %syb_api){
-	DBI->_install_method("DBI::$syb_api{$method}->[0]::$method", 
-			     'Sybase.pm',
-			     $syb_api{$method}->[1]);
-    }
-}
-
     sub driver {
 	return $drh if $drh;
 	my($class, $attr) = @_;
@@ -56,9 +50,27 @@
 	    'State'   => \$DBD::Sybase::sqlstate,
 	    'Attribution' => 'Sybase DBD by Michael Peppler',
 	    });
+
+	if($DBI::VERSION >= 1.34 && !$DBD::Sybase::init_done) {
+	    DBD::Sybase::db->install_method('syb_nsql');
+	    DBD::Sybase::db->install_method('syb_date_fmt');
+	    DBD::Sybase::st->install_method('syb_ct_get_data');
+	    DBD::Sybase::st->install_method('syb_ct_data_info');
+	    DBD::Sybase::st->install_method('syb_ct_send_data');
+	    DBD::Sybase::st->install_method('syb_ct_prepare_send');
+	    DBD::Sybase::st->install_method('syb_ct_finish_send');
+	    DBD::Sybase::st->install_method('syb_output_params');
+	    DBD::Sybase::st->install_method('syb_describe');
+	    ++$DBD::Sybase::init_done;
+	}
+
 	$drh;
     }
 
+    sub CLONE {
+	undef $drh;
+    }
+
 
     1;
 }
@@ -81,8 +93,53 @@
 
 	DBD::Sybase::db::_login($this, $server, $user, $auth, $attr) or return undef;
 
+	# OK - let us see what sort of server we're connected to. We
+	# need this because some of the SQL commands that we need to send
+	# to the server depend on this information.
+    {
+	
+	# local $this->{AutoCommit} = 1; #commented out - see bug #551.
+
+        # Attributes that are passed in only get set *after* this routine
+        # exits (i.e. when connect() returns).
+        # So we need to set any syb_xxx attributes here:
+        foreach (grep(/^syb/, keys(%$attr))) {
+	    $this->{$_} = $attr->{$_};
+	}
+	
+	my $sth = $this->prepare("select \@\@version");
+	if($sth->execute) {
+	    my $row = $sth->fetch;
+	    if($row && ref($row)) {
+		if($row->[0] =~ /microsoft/i) {
+		    $this->{syb_server_version} = -1;
+		    #warn "Connected to MS-SQL\n";
+		} elsif($row->[0] =~ /adaptive server enterprise\/([\d\.]+)/i ||
+		       $row->[0] =~ /sql server\/([\d\.]+)/i) {
+		    $this->{syb_server_version} = $1;
+		    #warn "Connected to Sybase\n";
+		}
+	    }
+	}
+    }
+
 	$this;
     }
+
+    sub data_sources {
+	my @s;
+	if ($^O eq 'MSWin32') {
+	    open(INTERFACES, "$ENV{SYBASE}/ini/sql.ini") or return;
+	    @s = map { /\[(\S+)\]/i; "dbi:Sybase:server=$1" } grep /\[/i, <INTERFACES>;
+	    close(INTERFACES);
+	} else {
+	    open(INTERFACES, "$ENV{SYBASE}/interfaces") or return;
+	    @s = map { /^(\S+)/i; "dbi:Sybase:server=$1" } grep /^[^\s\#]/i, <INTERFACES>;
+	    close(INTERFACES);
+	}
+
+	return @s;
+    }
 }
 
 
@@ -110,8 +167,25 @@
 
     sub tables {
 	my $dbh = shift;
+ 	my $catalog = shift;
+ 	my $schema = shift || '%';
+ 	my $table = shift || '%';
+ 	my $type = shift || '%';
+ 	$type =~ s/[\'\"\s]//g; # strip quotes and spaces
+ 	if ($type =~ /,/) { # multiple types
+	    $type = '[' . join('', map { substr($_,0,1) } split /,/, $type) . ']';
+ 	} else {
+ 		$type = substr($type,0,1);
+ 	}
+ 	$type =~ s/T/U/;
+  
+ 	my $sth;
+ 	if ($catalog and $catalog ne '%') {
+	    $sth = $dbh->prepare("select o.name from $catalog..sysobjects o, $catalog..sysusers u where o.type like '$type' and o.name like '$table' and o.uid = u.uid and u.name like '$schema'");
+ 	} else {
+	    $sth = $dbh->prepare("select o.name from sysobjects o, sysusers u where o.type like '$type' and o.name like '$table' and o.uid = u.uid and u.name like '$schema'");
+ 	}
 
-	my $sth = $dbh->prepare("select name from sysobjects where type in ('V', 'U')");
 	$sth->execute;
 	my @names;
 	my $dat;
@@ -131,7 +205,6 @@
 	my $sth = $dbh->prepare($statement, $attr) or return undef;
 	$sth->execute(@params) or return undef;
 	return undef if $sth->err;
-	my $rows = $sth->rows;
 	if(defined($sth->{syb_more_results})) {
 	    do {
 		while(my $dat = $sth->fetch) {
@@ -140,10 +213,29 @@
 		}
 	    } while($sth->{syb_more_results});
 	}
+	my $rows = $sth->rows;
 
 	($rows == 0) ? "0E0" : $rows;
     }
 
+    # This will only work if the statement handle used to do the insert
+    # has been properly freed. Otherwise this will try to fetch @@identity
+    # from a different (new!) connection - which is obviously wrong.
+    sub last_insert_id {
+	my ($dbh, $catalog, $schema, $table, $field, $attr) = @_;
+	# parameters are ignored.
+	
+	my $sth = $dbh->prepare('select @@identity');
+	if(!$sth->execute) {
+	    return undef;
+	}
+	my $value;
+	($value) = $sth->fetchrow_array;
+	$sth->finish;
+	
+	return $value;
+    }
+
     sub table_info {
 	my $dbh = shift;
 	my $catalog = $dbh->quote(shift);
@@ -227,6 +319,12 @@
 	$sth;
     }
 
+    sub begin_work {
+        my $dbh = shift;
+
+	return $dbh->STORE('AutoCommit', 0);
+    }
+
     sub ping {
 	my $dbh = shift;
 	return 0 if DBD::Sybase::db::_isdead($dbh);
@@ -250,37 +348,13 @@
 # Calling sp_datatype_info returns the appropriate data for the server that
 # we are currently connected to.
 # In general the data is static, so it's not really necessary, but ASE 12.5
-# introduces some changes, in particular char/varchar max lenghts of 2048...
-#	my $sth = $dbh->prepare("sp_datatype_info");
+# introduces some changes, in particular char/varchar max lenghts that depend
+# on the server's page size. 12.5.1 introduces the DATE and TIME datatypes.
+	my $sth = $dbh->prepare("sp_datatype_info");
 	my $data;
-#	if($sth->execute) {
-#	    $data = $sth->fetchall_arrayref;
-#	} else {
-	    $data = [
-       ['bit',-7,1,undef,undef,undef,'0','0',2,undef,'0',undef,'bit',undef,undef,undef,undef,undef,undef],
-       ['tinyint',-6,3,undef,undef,undef,1,'0',2,1,'0','0','tinyint',undef,undef,undef,undef,undef,undef],
-       ['image',-4,'2147483647','0x',undef,undef,1,'0',1,undef,'0',undef,'image',undef,undef,undef,undef,undef,undef],
-       ['timestamp',-3,8,'0x',undef,undef,1,'0',2,undef,'0',undef,'timestamp',undef,undef,undef,undef,undef,undef],
-       ['varbinary',-3,255,'0x',undef,'maxlength',1,'0',2,undef,'0',undef,'varbinary',undef,undef,undef,undef,undef,undef],
-       ['binary',-2,255,'0x',undef,'length',1,'0',2,undef,'0',undef,'binary',undef,undef,undef,undef,undef,undef],
-       ['text',-1,'2147483647','\'','\'',undef,1,1,1,undef,'0',undef,'text',undef,undef,undef,undef,undef,undef],
-       ['char',1,255,'\'','\'','length',1,1,3,undef,'0',undef,'char',undef,undef,undef,undef,undef,undef],
-       ['nchar',1,255,'\'','\'',undef,1,1,3,undef,'0',undef,'nchar',undef,undef,undef,undef,undef,undef],
-       ['numeric',2,38,undef,undef,'precision,scale',1,'0',2,'0','0','0','numeric','0',38,undef,undef,undef,undef],
-       ['decimal',3,38,undef,undef,'precision,scale',1,'0',2,'0','0','0','decimal','0',38,undef,undef,undef,undef],
-       ['money',3,19,'\$',undef,undef,1,'0',2,'0',1,'0','money',undef,undef,undef,undef,undef,undef],
-       ['smallmoney',3,10,'\$',undef,undef,1,'0',2,'0',1,'0','smallmoney',undef,undef,undef,undef,undef,undef],
-       ['int',4,10,undef,undef,undef,1,'0',2,'0','0','0','int',undef,undef,undef,undef,undef,undef],
-       ['smallint',5,5,undef,undef,undef,1,'0',2,'0','0','0','smallint',undef,undef,undef,undef,undef,undef],
-       ['float',6,15,undef,undef,undef,1,'0',2,'0','0','0','float',undef,undef,undef,undef,10,undef],
-       ['real',7,7,undef,undef,undef,1,'0',2,'0','0','0','real',undef,undef,undef,undef,10,undef],
-       ['datetime',11,23,'\'','\'',undef,1,'0',3,undef,'0',undef,'datetime',undef,undef,93,undef,undef,undef],
-       ['smalldatetime',11,16,'\'','\'',undef,1,'0',3,undef,'0',undef,'smalldatetime',undef,undef,93,undef,undef,undef],
-       ['nvarchar',12,255,'\'','\'',undef,1,1,3,undef,'0',undef,'nvarchar',undef,undef,undef,undef,undef,undef],
-       ['sysname',12,30,'\'','\'','maxlength',1,1,3,undef,'0',undef,'sysname',undef,undef,undef,undef,undef,undef],
-       ['varchar',12,255,'\'','\'','maxlength',1,1,3,undef,'0',undef,'varchar',undef,undef,undef,undef,undef,undef]
-		    ];
-#	}
+	if($sth->execute) {
+	    $data = $sth->fetchall_arrayref;
+	}
 	my $ti = 
 	[     {   TYPE_NAME         => 0,
 		  DATA_TYPE         => 1,
@@ -303,9 +377,15 @@
 		  interval_precision => 18,
 	      },
 	];
-#	foreach (@$data) {
-#	    push(@$ti, $_);
-#	}
+	# ASE 11.x only returns 13 columns:
+	my $c;
+	if(($c = scalar(@{$data->[0]})) < 19) {
+	  foreach (keys(%{$ti->[0]})) {
+	    if($ti->[0]->{$_} >= $c) {
+	      delete($ti->[0]->{$_});
+	    }
+	  }
+	}
 	push(@$ti, @$data);
 
 	return $ti;
@@ -320,6 +400,7 @@
 	my $retrycount = $dbh->FETCH('syb_deadlock_retry');
 	my $retrysleep = $dbh->FETCH('syb_deadlock_sleep') || 60;
 	my $retryverbose = $dbh->FETCH('syb_deadlock_verbose');
+	my $nostatus = $dbh->FETCH('syb_nsql_nostatus');
 
 #	warn "retrycount = $retrycount, retrysleep = $retrysleep, verbose = $retryverbose\n";
 
@@ -348,13 +429,24 @@
 	
     DEADLOCK: 
         {	
+	    # Initialize $err before each iteration through this loop.
+	    # Otherwise, we inherit the value from the previous failure.
+
+	    $err = undef;
+
 	    # Use RaiseError technique to throw a fatal error if anything goes
 	    # wrong in the execute or fetch phase.
 	    eval { 
 		$sth->execute || die $sth->errstr; 
 #		$sth->execute;
 		do {
-		    if ( $type eq "HASH" ) {
+		    if ( $DBD::Sybase::syb_nsql_nostatus &&
+			 $sth->{syb_result_type} == &CS_STATUS_RESULT ) {
+			while ( $data = $sth->fetchrow_arrayref ) {
+			    ; # do not include return status rows..
+			}  
+		    } 
+		    elsif ( $type eq "HASH" ) {
 			while ( $data = $sth->fetchrow_hashref ) {
 			    die $sth->errstr if($sth->err);
 			    if ( ref $callback eq "CODE" ) {
@@ -429,6 +521,10 @@
 	    return @res;
 	}
     }
+    
+    if($DBI::VERSION >= 1.34) {
+	*syb_nsql = *nsql;
+    }
 }
 
 
@@ -481,6 +577,13 @@
 	$status;
     }
 	
+#    if($DBI::VERSION >= 1.34) {
+#	*syb_ct_get_data = *ct_get_data;
+#	*syb_ct_data_info = *ct_data_info;
+#	*syb_ct_send_data = *ct_send_data;
+#	*syb_ct_prepare_send = *ct_prepare_send;
+#	*syb_ct_finish_send = *ct_finish_send;
+#    }
 		    
 }
 
@@ -524,7 +627,7 @@ C<BEGIN{}> block:
        $ENV{SYBASE} = '/opt/sybase/11.0.2';
    }
 
-   $dbh = DBI->connect('dbi:Sybase:', $user, $passwd);
+   my $dbh = DBI->connect('dbi:Sybase:', $user, $passwd);
 
 
 =head2 Specifying the server name
@@ -550,7 +653,7 @@ properties. Currently the following are supported:
 
 =item server
 
-Specify the server that we should connect to
+Specify the server that we should connect to.
 
      $dbh = DBI->connect("dbi:Sybase:server=BILLING",
 			 $user, $passwd);
@@ -558,6 +661,29 @@ Specify the server that we should connect to
 The default server is I<SYBASE>, or the value of the I<$DSQUERY> environment
 variable, if it is set.
 
+=item host
+=item port
+
+If you built DBD::Sybase with OpenClient 12.5.1 or later, then you can
+use the I<host> and I<port> values to define the server you want to
+connect to. This will by-pass the server name lookup in the interfaces file.
+This is useful in the case where the server hasn't been entered in the 
+interfaces file.
+
+     $dbh = DBI->connect("dbi:Sybase:host=db1.domain.com;port=4100",
+			 $user, $passwd);
+
+=item maxConnect
+
+By default DBD::Sybase (and the underlying OpenClient libraries) is limited
+to openening 25 simultaneous connections to one or more database servers.
+If you need more than 25 connections at the same time, you can use the
+I<maxConnect> option to increase this number.
+
+     $dbh = DBI->connect("dbi:Sybase:maxConnect=100",
+			 $user, $passwd);
+
+
 =item database
 
 Specify the database that should be made the default database.
@@ -635,14 +761,14 @@ timeout value. See also the Open Client documentation on CS_TIMEOUT.
 Specify the name for this connection that will be displayed in sp_who
 (ie in the sysprocesses table in the I<program_name> column).
 
-    $dbh->DBI->connect("dbi:Sybase:scriptName=myScript", $user, $password);
+    $dbh=DBI->connect("dbi:Sybase:scriptName=myScript", $user, $password);
 
 =item hostname
 
 Specify the hostname that will be displayed by sp_who (and will be stored
 in the hostname column of sysprocesses)..
 
-    $dbh->DBI->connect("dbi:Sybase:hostname=kiruna", $user, $password);
+    $dbh=DBI->connect("dbi:Sybase:hostname=kiruna", $user, $password);
 
 =item tdsLevel
 
@@ -652,7 +778,7 @@ In general this is automatically negotiated between the client and the
 server, but in certain cases this may need to be forced to a lower level
 by the client. 
 
-    $dbh->DBI->connect("dbi:Sybase:tdsLevel=CS_TDS_42", $user, $password);
+    $dbh=DBI->connect("dbi:Sybase:tdsLevel=CS_TDS_42", $user, $password);
 
 B<NOTE>: Setting the tdsLevel below CS_TDS_495 will disable a number of
 features, ?-style placeholders and CHAINED non-AutoCommit mode, in particular.
@@ -662,7 +788,47 @@ features, ?-style placeholders and CHAINED non-AutoCommit mode, in particular.
 Specify the use of the client password encryption supported by CT-Lib.
 Specify a value of 1 to use encrypted passwords.
 
-    $dbh->DBI->connect("dbi:Sybase:encryptPassword=1", $user, $password);
+    $dbh=DBI->connect("dbi:Sybase:encryptPassword=1", $user, $password);
+
+=item kerberos
+
+Note: Requires OpenClient 11.1.1 or later.
+
+Sybase and OpenClient can use Kerberos to perform network-based login.
+If you use Kerberos for authentication you can use this feature and pass
+a kerberos serverprincipal using the C<kerberos=value> parameter:
+
+    $dbh = DBI->connect("dbi:Sybase:kerberos=$serverprincipal", '', '');
+
+In addition, if you have a system for retrieving Kerberos serverprincipals at
+run-time you can tell DBD::Sybase to call a perl subroutine to get
+the serverprincipal from connect():
+
+    sub sybGetPrinc {
+        my $srv = shift;
+        return the serverprincipal...
+    }
+    $dbh = DBI->connect('dbi:Sybase:server=troll', '', '', { syb_kerberos_serverprincipal => \&sybGetPrinc });
+
+The subroutine will be called with one argument (the server that we will
+connect to, using the normal Sybase behavior of checking the DSQUERY
+environment variable if no server is specified in the connect()) and is
+expected to return a string (the Kerberos serverprincipal) to the caller.
+
+=item sslCAFile
+
+Specify the location of an alternate I<trusted.txt> file for SSL
+connection negotiation:
+
+  $dbh->DBI->connect("dbi:Sybase:sslCAFile=/usr/local/sybase/trusted.txt.ENGINEERING", $user, $password); 
+
+=item bulkLogin
+
+Set this to 1 if the connection is going to be used for a bulk-load
+operation (see I<Experimental Bulk-Load functionality> elsewhere in this
+document.)
+
+  $dbh->DBI->connect("dbi:Sybase:bulkLogin=1", $user, $password); 
 
 =back
 
@@ -672,7 +838,6 @@ together by separating each entry with a semi-colon:
     $dbh = DBI->connect("dbi:Sybase:server=ENGINEERING;packetSize=8192;language=us_english;charset=iso_1",
 			$user, $pwd);
 
-
 =head1 Handling Multiple Result Sets
 
 Sybase's Transact SQL has the ability to return multiple result sets
@@ -731,13 +896,11 @@ to your script then you can use the symbolic values (CS_xxx_RESULT)
 instead of the numeric values in your programs, which should make them 
 easier to read.
 
-See also the C<syb_output_param> func() call to handle stored procedures 
+See also the C<$sth->syb_output_params> call to handle stored procedures 
 that B<only> return B<OUTPUT> parameters.
 
 =head1 $sth->execute() failure mode behavior
 
-B<THIS HAS CHANGED IN VERSION 0.21!>
-
 DBD::Sybase has the ability to handle multi-statement SQL commands
 in a single batch. For example, you could insert several rows in 
 a single batch like this:
@@ -902,7 +1065,8 @@ not supported when using DBD::Sybase to connect to a MS-SQL server.
 =item syb_chained_txn (bool)
 
 If set then we use CHAINED transactions when AutoCommit is off. 
-Otherwise we issue an explicit BEGIN TRAN as needed. The default is off.
+Otherwise we issue an explicit BEGIN TRAN as needed. The default is on
+if it is supported by the server.
 
 This attribute should usually be used only during the connect() call:
 
@@ -992,8 +1156,6 @@ Default: off
 
 =item syb_bind_empty_string_as_null (bool)
 
-B<New in 0.95>
-
 If this attribute is set then an empty string (i.e. "") passed as
 a parameter to an $sth->execute() call will be converted to a NULL
 value. If the attribute is not set then an empty string is converted to
@@ -1003,8 +1165,6 @@ Default: off
 
 =item syb_cancel_request_on_error (bool)
 
-B<New in 0.95>
-
 If this attribute is set then a failure in a multi-statement request
 (for example, a stored procedure execution) will cause $sth->execute()
 to return failure, and will cause any other results from this request to 
@@ -1015,6 +1175,12 @@ up to version 0.94.
 
 Default: on
 
+=item syb_date_fmt (string)
+
+Defines the date/time conversion string when fetching data. See the 
+entry for the C<syb_date_fmt()> method elsewhere in this document for a
+description of the available formats.
+
 =back
 
 =head2 Statement Handle Attributes
@@ -1045,13 +1211,13 @@ when calling $sth->fetch (or any variation).
 
 Instead, you would use
 
-    $sth->func($column, \$data, $size, 'ct_get_data');
+    $sth->syb_ct_get_data($column, \$data, $size);
 
 to retrieve the IMAGE or TEXT data. If $size is 0 then the entire item is
 fetched, otherwis  you can call this in a loop to fetch chunks of data:
 
     while(1) {
-	$sth->func($column, \$data, 1024, 'ct_get_data');
+        $sth->syb_ct_get_data($column, \$data, 1024);
 	last unless $data;
 	print OUT $data;
     }
@@ -1075,12 +1241,11 @@ elsewhere in this document.
 
 By default DBD::Sybase will return I<DATETIME> and I<SMALLDATETIME>
 columns in the I<Nov 15 1998 11:13AM> format. This can be changed
-via a special B<_date_fmt()> function that is accessed via the $dbh->func()
-method.
+via a private B<syb_date_fmt()> method.
 
 The syntax is
 
-    $dbh->func($fmt, '_date_fmt');
+    $dbh->syb_date_fmt($fmt);
 
 where $fmt is a string representing the format that you want to apply.
 
@@ -1121,6 +1286,20 @@ Nov 15 1998 11:30AM
 
 11:30:11
 
+=item ISO
+
+2004-08-21 14:36:48.080
+
+=item ISO_strict
+
+2004-08-21T14:36:48.080Z
+
+Note that Sybase has no concept of a timezone, so the trailing "Z" is
+really not correct (assumes that the time is in UTC). However, there
+is no guarantee that the client and the server run in the same timezone,
+so assuming the timezone of the client isn't really a valid option
+either.
+
 =back
 
 =head1 Retrieving OUTPUT parameters from stored procedures
@@ -1142,7 +1321,7 @@ can use this shorthand:
 
     $sth = $dbh->prepare('...');
     $sth->execute;
-    @results = $sth->func('syb_output_params');
+    @results = $sth->syb_output_params();
 
 This will return an array for all the OUTPUT parameters in the proc call,
 and will ignore any other results. The array will be undefined if there are 
@@ -1206,7 +1385,16 @@ There is an alternative way to access and update IMAGE/TEXT data
 using the natice OpenClient API. This is done via $h->func() calls,
 and is, unfortunately, a little convoluted.
 
-=head2 Handling IMAGE/TEXT data with ct_get_data()/ct_send_data()
+=head2 Handling IMAGE/TEXT data with syb_ct_get_data()/syb_ct_send_data()
+
+With DBI 1.34 and later you can call all of these ct_xxx() calls directly
+as statement handle methods by prefixing them with syb_, so for example
+
+    $sth->func($col, $dataref, $numbytes, 'ct_fetch_data');
+
+becomes
+
+    $sth->syb_ct_fetch_data($col, $dataref, $numbytes);
 
 =over 4
 
@@ -1230,20 +1418,23 @@ The call sequence is:
     $sth->execute;
     while($d = $sth->fetchrow_arrayref) {
        # The data is in the second column
-       $len = $sth->func(2, \$img, 0, 'ct_get_data');
+       $len = $sth->syb_ct_get_data(2, \$img, 0);
+       # with DBI 1.33 and earlier, this would be
+       # $len = $sth->func(2, \$img, 0, 'ct_get_data');
     }
 
 ct_get_data() returns the number of bytes that were effectively fetched,
 so that when fetching chunks you can do something like this:
 
    while(1) {
-      $len = $sth->func(2, $imgchunk, 1024, 'ct_get_data');
+      $len = $sth->syb_ct_get_data(2, $imgchunk, 1024);
       ... do something with the $imgchunk ...
       last if $len != 1024;
    }
 
 To explain further: Sybase stores IMAGE/TEXT data separately from 
-normal table data, in a chain of 2k blocks. To update an IMAGE/TEXT
+normal table data, in a chain of pagesize blocks (a Sybase database page
+is defined at the server level, and can be 2k, 4k, 8k or 16k in size.) To update an IMAGE/TEXT
 column Sybase needs to find the head of this chain, which is known as
 the "text pointer". As there is no I<where> clause when the ct_send_data()
 API is used we need to retrieve the I<text pointer> for the correct
@@ -1291,24 +1482,25 @@ up via ct_prepare_send() and ct_data_info() for this to work. ct_send_data()
 returns B<TRUE> on success, and B<FALSE> on failure.
 
 In this example, we wish to update the data in the I<img> column
-where the I<id> column is 1:
+where the I<id> column is 1. We assume that DBI is at version 1.34 or
+later and use the direct method calls:
 
   # first we need to find the CS_IODESC data for the data
   $sth = $dbh->prepare("select img from imgtable where id = 1");
   $sth->execute;
   while($sth->fetch) {    # don't care about the data!
-      $sth->func('CS_GET', 1, 'ct_data_info');
+      $sth->syb_ct_data_info('CS_GET', 1);
   }
 
   # OK - we have the CS_IODESC values, so do the update:
-  $sth->func('ct_prepare_send');
+  $sth->syb_ct_prepare_send();
   # Set the size of the new data item (that we are inserting), and make
   # the operation unlogged
-  $sth->func('CS_SET', 1, {total_txtlen => length($image), log_on_update => 0}, 'ct_data_info');
+  $sth->syb_ct_data_info('CS_SET', 1, {total_txtlen => length($image), log_on_update => 0});
   # now transfer the data (in a single chunk, this time)
-  $sth->func($image, length($image), 'ct_send_data');
+  $sth->syb_ct_send_data($image, length($image));
   # commit the operation
-  $sth->func('ct_finish_send');
+  $sth->ct_finish_send();
 
 The ct_send_data() call can also transfer the data in chunks, however you 
 must know the total size of the image before you start the insert. For example:
@@ -1319,14 +1511,14 @@ must know the total size of the image before you start the insert. For example:
   $sth = $dbh->prepare("select img from imgtable where id = 1");
   $sth->execute;
   while($sth->fetch) {    # don't care about the data!
-      $sth->func('CS_GET', 1, 'ct_data_info');
+      $sth->syb_ct_data_info('CS_GET', 1);
   }
 
   # OK - we have the CS_IODESC values, so do the update:
-  $sth->func('ct_prepare_send');
+  $sth->syb_ct_prepare_send();
   # Set the size of the new data item (that we are inserting), and make
   # the operation unlogged
-  $sth->func('CS_SET', 1, {total_txtlen => $size, log_on_update => 0}, 'ct_data_info');
+  $sth->syb_ct_data_info('CS_SET', 1, {total_txtlen => $size, log_on_update => 0});
 
   # open the file, and store it in the db in 1024 byte chunks.
   open(IN, $file) || die "Can't open $file: $!";
@@ -1335,11 +1527,11 @@ must know the total size of the image before you start the insert. For example:
       $bytesread = read(IN, $buff, $to_read);
       $size -= $bytesread;
 
-      $sth->func($buff, $bytesread, 'ct_send_data');
+      $sth->syb_ct_send_data($buff, $bytesread);
   }
   close(IN);
   # commit the operation
-  $sth->func('ct_finish_send');
+  $sth->syb_ct_finish_send();
       
 
 =back
@@ -1370,10 +1562,25 @@ be held on certain system tables for the duration of the transaction.
 If $h->{syb_chained_txn} is I<on>, then DBD::Sybase sets the
 I<CHAINED> option, which tells Sybase not to commit anything automatically.
 Again, you will need to call $dbh->commit() to make any changes to the data
-permanent. In this case Sybase will not let you issue I<BEGIN TRAN> 
-statements in the SQL code that is executed, so if you need to execute
-stored procedures that have I<BEGIN TRAN> statements in them you 
-must use $h->{syb_chained_txn} = 0, or $h->{AutoCommit} = 1.
+permanent. 
+
+=head2 Behavior of $dbh->last_insert_id
+
+This version of DBD::Sybase includes support for the last_insert_id() call,
+with the following caveats:
+
+The last_insert_id() call is simply a wrapper around a "select @@identity"
+query. To be successful (i.e. to return the correct value) this must
+be executed on the same connection as the INSERT that generated the
+new IDENTITY value. Therefore the statement handle that was used to
+perform the insert B<must> have been closed/freed before last_insert_id()
+can be called. Otherwise last_insert_id() will be forced to open a different
+connection to perform the query, and will return an invalid value (usually
+in this case it will return 0).
+
+last_insert_id() ignores any parameters passed to it, and will NOT return
+the last @@identity value generated in the case where placeholders were used,
+or where the insert was encapsulated in a stored procedure.
 
 =head1 Using ? Placeholders & bind parameters to $sth->execute
 
@@ -1429,7 +1636,7 @@ The DBI docs mention the following regarding NULL values and placeholders:
 
 =back
 
-This will not work with a Sybase database server. If you attempt the 
+This will I<not> work with a Sybase database server. If you attempt the 
 above construct you will get the following error:
 
 =over 4
@@ -1517,9 +1724,7 @@ at http://sybooks.sybase.com/
 
 =head1 Stored Procedures and Placeholders
 
-B<NOTE: This feature is experimental>
-
-This version of DBD::Sybase introduces the ability to use ?-style
+DBD::Sybase has the ability to use ?-style
 placeholders as parameters to stored proc calls. The requirements are
 that the stored procedure call be initiated with an "exec" and that it be
 the only statement in the batch that is being prepared():
@@ -1536,13 +1741,20 @@ You can also use positional parameters:
 
 You may I<not> mix positional and named parameter in the same prepare.
 
+You I<can't> mix placeholder parameters and hard coded parameters. For example
+
+    $sth = $dbh->prepare("exec my_proc \@p1 = 1, \@p2 = ?");
+
+will I<not> work - because the @p1 parameter isn't parsed correctly
+and won't be sent to the server.
+
 You can specify I<OUTPUT> parameters in the usual way, but you can B<NOT>
 use bind_param_inout() to get the output result - instead you have to call
 fetch() and/or $sth->func('syb_output_params'):
 
     my $sth = $dbh->prepare("exec my_proc \@p1 = ?, \@p2 = ?, \@p3 = ? OUTPUT ");
     $sth->execute('one', 'two', 'three');
-    my (@data) = $sth->func('syb_output_params');
+    my (@data) = $sth->syb_output_params();
 
 DBD::Sybase does not attempt to figure out the correct parameter type
 for each parameter (it would be possible to do this for most cases, but
@@ -1585,26 +1797,137 @@ setting the I<arithabort> option:
 See the I<set> command in the Sybase Adaptive Server Enterprise Reference 
 Manual for more information on the set command and on the arithabort option.
 
+=head1 Other Private Methods
 
-=head1 BUGS
+=head2 DBD::Sybase private Statement Handle Methods
 
-You can run out of space in the tempdb database if you use a lot of
-calls with bind variables (ie ?-style placeholders) without closing the
-connection. On my system, with an 8 MB tempdb database I run out of space
-after 760 prepare() statements with ?-parameters. This is because
-Sybase creates stored procedures for each prepare() call. So my
-suggestion is to only use ?-style placeholders if you really need them
-(i.e. if you are going to execute the same prepared statement multiple
-times).
+=over 4
 
-The new primary_key_info() method will only return data for tables 
-where a declarative "primary key" constraint was included when the table
-was created.
+=item @data = $sth->syb_describe([$assoc])
 
-I have a simple bug tracking database at http://www.peppler.org/cgi-bin/bug.cgi .
-You can use it to view known problems, or to report new ones. Keep in
-mind that peppler.org is connected to the net via a K56 dialup line, so
-it may be slow.
+Retrieves the description of each of the output columns of the current 
+result set. Each element of the returned array is a reference
+to a hash that describes the column. The following fields are set:
+NAME, TYPE, SYBTYPE, MAXLENGTH, SCALE, PRECISION, STATUS.
+
+You could use it like this:
+
+   my $sth = $dbh->prepare("select name, uid from sysusers");
+   $sth->execute;
+   my @description = $sth->syb_describe;
+   print "$description[0]->{NAME}\n";         # prints ’name’
+   print "$description[0]->{MAXLENGTH}\n";    # prints 30
+   ....
+
+   while(my $row = $sth->fetch) {
+      ....
+   }
+
+The STATUS field is a string which can be tested for the following
+values: CS_CANBENULL, CS_HIDDEN, CS_IDENTITY, CS_KEY, CS_VERSION_KEY, 
+CS_TIMESTAMP and CS_UPDATABLE. See table 3-46 of the Open Client Client 
+Library Reference Manual for a description of each of these values.
+
+The TYPE field is the data type that Sybase::CTlib converts the
+column to when retrieving the data, so a DATETIME column will be
+returned as a CS_CHAR_TYPE column.
+
+The SYBTYPE field is the real Sybase data type for this column.
+
+I<Note that the symbolic values of the CS_xxx symbols isn't available
+yet in DBD::Sybase.>
+
+
+=back
+
+=head1 Experimental Bulk-Load Functionality
+
+Starting with release 1.04.2 DBD::Sybase has the ability to use Sybase's
+BLK (bulk-loading) API to perform fast data loads. Basic usage is as follows:
+
+  my $dbh = DBI->connect('dbi:Sybase:server=MY_SERVER;bulkLogin=1', $user, $pwd);
+
+  $dbh->begin_work;  # optional.
+  my $sth = $dbh->prepare("insert the_table values(?, ?, ?, ?, ?)",
+                          {syb_bcp_attribs => { identity_flag => 0,
+                                               identity_column => 0 }}});
+  while(<DATA>) {
+    chomp;
+    my @row = split(/\|/, $_);   # assume a pipe-delimited file...
+    $sth->execute(@row);
+  }
+  $dbh->commit;
+  print "Sent ", $sth->rows, " to the server\n";
+  $sth->finish;
+
+First, you need to specify the new I<bulkLogin> attribute in the connection
+string, which turns on the CS_BULK_LOGIN property for the connection. Without
+this property the BLK api will not be functional.
+
+You call $dbh->prepare() with a regular INSERT statement and the 
+special I<syb_bcp_attribs> attribute to turn on BLK handling of the data.
+The I<identity_flag> sub-attribute can be set to 1 if your source data
+includes the values for the target table's IDENTITY column. If the
+target table has an IDENTITY column but you want the insert operation to
+generate a new value for each row then leave I<identity_flag> at 0, but set
+I<identity_col> to the column number of the identity column (it's usually
+the first column in the table, but not always.)
+
+The number of placeholders in the INSERT statement I<must> correspond to
+the number of columns in the table, and the input data I<must> be in the
+same order as the table's physical column order. Any column list in the
+INSERT statement (i.e. I<insert table(a, b, c,...) values(...)> is ignored.
+
+The value of AutoCommit is ignored for BLK operations - rows are only 
+commited when you call $dbh->commit.
+
+You can call $dbh->rollback to cancel any uncommited rows, but this I<also>
+cancels the rest of the BLK operation: any attempt to load rows to the
+server after a call to $dbh->rollback() will fail.
+
+If a row fails to load due to a CLIENT side error (such as a data conversion
+error) then $sth->execute() will return a failure (i.e. false) and
+$sth->errstr will have the reason for the error.
+
+If a row fails on the SERVER side (for example due to a duplicate row
+error) then the entire batch (i.e. between two $dbh->commit() calls) 
+will fail. This is normal behavior for BLK/bcp.
+
+The Bulk-Load API is very sensitive to data conversion issues, as all the
+conversions are handled on the client side, and the row is pre-formatted
+before being sent to the server. By default any conversion that is flagged
+by Sybase's cs_convert() call will result in a failed row. Some of these
+conversion errors are patently fatal (e.g. converting 'Feb 30 2001' to a
+DATETIME value...), while others are debatable (e.g. converting 123.456 to
+a NUMERIC(6,2) which results in a loss of precision). The default behavior
+of failing any row that has a conversion error in it can be modified by 
+using a special error handler. Returning 0 from this handler
+tells DBD::Sybase to fail this row, and returning 1 means that we still
+want to try to send the row to the server (obviously Sybase's internal
+code can still fail the row at that point.)
+You set the handler like this:
+
+    DBD::Sybase::syb_set_cslib_cb(\&handler);
+
+and a sample handler:
+
+   sub cslib_handler {
+     my ($layer, $origin, $severity, $errno, $errmsg, $osmsg, $blkmsg) = @_;
+     
+     print "Layer: $layer, Origin: $origin, Severity: $severity, Error: $errno\n";
+     print $msg;
+     print $osmsg if($osmsg);
+     print $blkmsg if $blkmsg;
+
+     return 1 if($errno == 36)
+
+     return 0;
+   }
+
+Please see the t/xblk.t test script for some examples.
+
+Reminder - this is an I<experimental> implementation. It may change
+in the future, and it could be buggy.
 
 =head1 Using DBD::Sybase with MS-SQL 
 
@@ -1635,6 +1958,10 @@ Usage:
 
    @data = $dbh->func($sql, $type, $callback, 'nsql');
 
+If the DBI version is 1.34 or later, then you can also call it this way:
+
+   @data = $dbh->syb_nsql($sql, $type, $callback);
+
 This executes the query in $sql, and returns all the data in @data. The 
 $type parameter can be used to specify that each returned row be in array
 form (i.e. $type passed as 'ARRAY', which is the default) or in hash form 
@@ -1646,7 +1973,7 @@ being returned by the routine (to allow processing of large result sets,
 for example).
 
 C<nsql> also checks three special attributes to enable deadlock retry logic
-\(I<Note> none of these attributes have any effect anywhere else at the moment):
+(I<Note> none of these attributes have any effect anywhere else at the moment):
 
 =over 4
 
@@ -1664,11 +1991,67 @@ Number of seconds to sleep between deadlock retries. Default is 60.
 
 Enable verbose logging of deadlock retry logic. Default is off.
 
+=item syb_nsql_nostatus (bool)
+
+If true then stored procedure return status values (i.e. results of type
+CS_STATUS_RESULT) are ignored.
+
 =back
 
 Deadlock detection will be added to the $dbh->do() method in a future
 version of DBD::Sybase. 
 
+=head1 Multi-Threading
+
+DBD::Sybase is thread-safe (i.e. can be used in a multi-threaded
+perl application where more than one thread accesses the database
+server) with the following restrictions:
+
+=over 4
+
+=item * perl version >= 5.8
+
+DBD::Sybase requires the use of I<ithreads>, available in the perl 5.8.0
+release. It will not work with the older 5.005 threading model.
+
+=item * Sybase thread-safe libraries
+
+Sybase's Client Library comes in two flavors. DBD::Sybase must find the
+thread-safe version of the libraries (ending in _r on Unix/linux). This 
+means Open Client 11.1.1 or later. In particular this means that you can't
+use the 10.0.4 libraries from the free 11.0.3.3 release on linux if you
+want to use multi-threading.
+
+=item * use DBD::Sybase
+
+You I<must> include the C<use DBD::Sybase;> line in your program. This
+is needed because DBD::Sybase needs to do some setup I<before> the first
+thread is started.
+
+=back
+
+You can check to see if your version of DBD::Sybase is thread-safe at
+run-time by calling DBD::Sybase::thread_enabled(). This will return
+I<true> if multi-threading is available.
+
+See t/thread.t for a simple example.
+
+=head1 BUGS
+
+You can run out of space in the tempdb database if you use a lot of
+calls with bind variables (ie ?-style placeholders) without closing the
+connection and Sybase 11.5.x or older. This is because
+Sybase creates stored procedures for each prepare() call. 
+In 11.9.x and later Sybase will create "light-weight" stored procedures
+which don't use up any space in the tempdb database.
+
+The primary_key_info() method will only return data for tables 
+where a declarative "primary key" constraint was included when the table
+was created.
+
+I have a simple bug tracking database at http://www.peppler.org/cgi-bin/bug.cgi .
+You can use it to view known problems, or to report new ones. 
+
 
 =head1 SEE ALSO
 
@@ -1684,11 +2067,9 @@ DBD::Sybase by Michael Peppler
 
 =head1 COPYRIGHT
 
-The DBD::Sybase module is Copyright (c) 1997-2001 Michael Peppler.
+The DBD::Sybase module is Copyright (c) 1996-2004 Michael Peppler.
 The DBD::Sybase module is free software; you can redistribute it and/or
-modify it under the same terms as Perl itself with the exception that it
-cannot be placed on a CD-ROM or similar media for commercial distribution
-without the prior approval of the author.
+modify it under the same terms as Perl itself.
 
 =head1 ACKNOWLEDGEMENTS
 
